local ORI_ConfigCache, max, min, abs, sin, cos, atan2 = {}, math.max, math.min, math.abs, sin, cos, atan2

local function cc(m, f, ...)
	f[m](f, ...)
	return f
end
local darken do
	local CSL = CreateFrame("ColorSelect")
	function darken(r,g,b, vf, sf)
		CSL:SetColorRGB(r,g,b)
		local h,s,v = CSL:GetColorHSV()
		CSL:SetColorHSV(h, s*(sf or 1), v*(vf or 1))
		return CSL:GetColorRGB()
	end
end
local function shortBindName(bind)
	local a, s, c, k = bind:match("ALT%-"), bind:match("SHIFT%-"), bind:match("CTRL%-"), bind:match("[^-]*.$"):gsub("^(.).-(%d+)$","%1%2");
	return (a and "A" or "") .. (s and "S" or "") .. (c and "C" or "") .. k;
end
local function cooldownFormat(cd)
	if cd == 0 or not cd then return "" end
	local f, n, unit = cd > 10 and "%d%s" or "%.1f", cd, ""
	if n > 86400 then n, unit = ceil(n/86400), "d"
	elseif n > 3600 then n, unit = ceil(n/3600), "h"
	elseif n > 60 then n, unit = ceil(n/60), "m"
	elseif n > 10 then n = ceil(n) end
	return f, n, unit
end

local ORI_Frame = cc("SetFrameStrata", cc("SetSize", CreateFrame("Frame", nil, UIParent), 128, 128), "FULLSCREEN")
ORI_Frame.anchor = cc("SetPoint", cc("SetSize", CreateFrame("Frame", nil, UIParent), 1, 1), "CENTER")
ORI_Frame:SetPoint("CENTER", ORI_Frame.anchor)
local ORI_SetRotationPeriod, CreateQuadTexture do
	local function qf(f)
		return function (self, ...)
			for i=1,4 do
				local v = self[i];
				v[f](v, ...);
			end
		end
	end
	local quad, animations, quadPoints, quadTemplate = {}, {}, {"BOTTOMRIGHT", "BOTTOMLEFT", "TOPLEFT", "TOPRIGHT"}, {__index={SetVertexColor=qf("SetVertexColor"), SetAlpha=qf("SetAlpha"), SetShown=qf("SetShown")}}
	for i=1,4 do
		quad[i] = cc("SetPoint", cc("SetSize", CreateFrame("Frame", nil, ORI_Frame), 32, 32), quadPoints[i], ORI_Frame, "CENTER");
		local g = cc("SetLooping", cc("SetIgnoreFramerateThrottle", quad[i]:CreateAnimationGroup(), 1), "REPEAT");
		animations[i] = cc("SetOrigin", cc("SetDegrees", cc("SetDuration", g:CreateAnimation("Rotation"), 4), -360), quadPoints[i], 0, 0);
		g:Play();
	end
	function CreateQuadTexture(parent, layer, size, file)
		local group, size = setmetatable({}, quadTemplate), size/2
		for i=1,4 do
			local tex, d, l = cc("SetSize", cc("SetTexture", (parent or quad[i]):CreateTexture(nil, layer), file), size, size), i > 2, 2 > i or i > 3
			tex:SetTexCoord(l and 0 or 1, l and 1 or 0, d and 1 or 0, d and 0 or 1)
			group[i] = cc("SetPoint", tex, quadPoints[i], parent or quad[i], parent and "CENTER" or quadPoints[i])
		end
		return group
	end
	function ORI_SetRotationPeriod(p)
		local p = max(0.1, p)
		for i=1,4 do animations[i]:SetDuration(p) end
	end
end
local ORI_Circle = CreateQuadTexture(nil, "ARTWORK", 64, [[Interface\AddOns\OPie\gfx\circle]])
local ORI_Glow = CreateQuadTexture(nil, "BACKGROUND", 128, [[Interface\AddOns\OPie\gfx\glow]])
local ORI_Pointer = cc("SetTexture", cc("SetPoint", cc("SetSize", ORI_Frame:CreateTexture(nil, "ARTWORK"), 192, 192), "CENTER"), [[Interface\AddOns\OPie\gfx\pointer]])
local ORI_CenterCaption = cc("SetPoint", ORI_Frame:CreateFontString(nil, "OVERLAY", "GameFontHighlightSmall"), "TOP", ORI_Frame, "CENTER", 0, -52)
local ORI_CenterCooldownText = cc("SetPoint", ORI_Frame:CreateFontString(nil, "OVERLAY", "NumberFontNormalHuge"), "CENTER")

local ORI_CreateIndicator, ORI_CreateDefaultIndicator do
	local api = {}
	function api:SetIcon(texture)
		self.icon:SetTexture(texture)
		local ofs = (texture:match("^[Ii][Nn][Tt][Ee][Rr][Ff][Aa][Cc][Ee][\\/][Ii][Cc][Oo][Nn][Ss][\\/]") or texture == [[Interface\AddOns\OPie\gfx\opie_ring_icon]]) and (2/64) or (-2/64)
		self.icon:SetTexCoord(ofs, 1-ofs, ofs, 1-ofs)
	end
	function api:SetIconTexCoord(a,b,c,d, e,f,g,h)
		if a and b and c and d then
			if e and f and g and h then
				self.icon:SetTexCoord(a,b,c,d, e,f,g,h)
			else
				self.icon:SetTexCoord(a,b,c,d)
			end
		end
	end
	function api:SetIconVertexColor(r,g,b)
		self.icon:SetVertexColor(r,g,b)
	end
	function api:SetDominantColor(r,g,b)
		r, g, b = r or 1, g or 1, b or 0.6
		local cd, r2, g2, b2 = self.cd, darken(r,g,b, 0.20)
		local r3, g3, b3 = darken(r,g,b, 0.10, 0.50)
		self.hiEdge:SetVertexColor(r, g, b)
		self.iglow:SetVertexColor(r, g, b)
		self.oglow:SetVertexColor(r, g, b)
		self.edge:SetVertexColor(darken(r,g,b, 0.80))
		self.cdText:SetTextColor(r, g, b)		
		cd.spark:SetVertexColor(r, g, b)
		for i=1,4 do
			cd[i]:SetVertexColor(r2, g2, b2)
			cd[i+4]:SetVertexColor(r3, g3, b3)
		end
		cd[9]:SetVertexColor(r3, g3, b3)
	end
	function api:SetOverlayIcon(texture, w, h, ...)
		if not texture then
			self.overIcon:Hide()
		else
			self.overIcon:Show()
			self.overIcon:SetTexture(texture)
			self.overIcon:SetSize(w, h)
			if ... then
				self.overIcon:SetTexCoord(...)
			end
		end
	end
	function api:SetCount(count)
		self.count:SetText(count or "")
	end
	function api:SetBindingText(text)
		self.key:SetText(text or "")
	end
	function api:SetCooldown(remain, duration, usable)
		if (duration or 0) <= 0 or (remain or 0) <= 0 then
			self.cd:Hide()
		else
			local expire, usable, cd = GetTime() + remain, not not usable, self.cd
			local d = expire - (cd.expire or 0)
			if d < -0.05 or d > 0.05 then
				cd.duration, cd.expire, cd.updateCooldownStep, cd.updateCooldown = duration, expire, duration/1536/self[0]:GetEffectiveScale()
				cd:Show()
			end
			if cd.usable ~= usable then
				cd.usable = usable
				for i=1,4 do cd[i]:SetAlpha(usable and 0.45 or 1) end
				for i=5,9 do cd[i]:SetAlpha(usable and 0.25 or 0.85) end
				cd.spark:SetShown(usable)
			end
		end
	end
	function api:SetCooldownFormattedText(format, ...)
		self.cdText:SetFormattedText(format, ...)
	end
	function api:SetHighlighted(highlight)
		self.hiEdge:SetShown(highlight)
	end
	function api:SetActive(active)
		self.iglow:SetShown(active)
	end
	function api:SetOuterGlow(shown)
		self.oglow:SetShown(shown)
	end
	for f in ("SetPoint SetScale GetScale SetShown SetParent"):gmatch("%a+") do
		api[f] = function(self, ...)
			local w = self[0]
			return w[f](w, ...)
		end
	end
	local createCooldown do
		local function onUpdate(self, elapsed)
			local ucd, expire, time = self.updateCooldown or 0, self.expire or 0, GetTime()
			if ucd > elapsed and time < expire then
				self.updateCooldown = ucd - elapsed
				return
			end
			self.updateCooldown = self.updateCooldownStep
			local duration = self.duration or 0
			local progress = 1 - (expire - time)/duration
			if progress > 1 or duration == 0 then
				self:Hide()
			else
				progress = progress < 0 and 0 or progress
				local tri, pos, sp, pp, scale = self[9], 1+4*(progress - progress % 0.25), progress % 0.25 >= 0.125, (progress % 0.125) * 8, self.scale
				if self.pos ~= pos then
					for i=1,4 do
						self[i]:SetShown(i >= pos)
						self[4+i]:SetShown(i > pos or (i == pos and not sp))
						if i > pos then
							self[i]:SetSize(24, 24)
							local L, T = i > 2, i == 1 or i == 4
							self[i]:SetTexCoord(L and 0 or 0.5, L and 0.5 or 1, T and 0 or 0.5, T and 0.5 or 1)
							self[4+i]:SetSize(21*scale, 21*scale)
						end
					end
					tri:ClearAllPoints()
					tri:SetPoint((pos % 4 < 2 and "BOTTOM" or "TOP") .. (pos < 3 and "LEFT" or "RIGHT") , self, "CENTER")
					local iH, iV = pos == 2 or pos == 3, pos > 2
					tri:SetTexCoord(iH and 1 or 0, iH and 0 or 1, iV and 1 or 0, iV and 0 or 1)
					self.pos = pos
				end

				local l, r, inv = sp and 21 or (pp * 21), 21 - (sp and pp * 21 or 0), pos == 2 or pos == 4
				l, r = l > 0 and l or 0.00000001, r > 0 and r or 0.00000001
				tri:SetSize((inv and r or l)*scale, (inv and l or r)*scale)

				local chunk, shrink = self[4+pos], 21 - 21*pp
				chunk:SetSize((inv and 21 or shrink)*scale, (inv and shrink or 21)*scale)
				chunk:SetShown(not sp or pp >= 0.99)

				local p1, p2, e, p1a, p2a = sp and 1 or pp, sp and pp or 0, self[pos]
				if p1 > 0.9 and p2 < 0.1 then
					p1a = 0.9 + (p1 + p2 - 0.9)/2
					p2a = 1-(1.81 - p1a*p1a)^0.5
				else
					p1a, p2a = p1, p2
				end
				if p2 > 0.1 then
					p2 = 0.14 + (p2 - 0.10)*0.86/0.9
				elseif p2 > 0 then
					p2 = p2 + 0.04
				elseif p1 > 0.96 then
					p1, p2 = 1, p1-0.96
				elseif p1 > 0.75 then
					p1 = p1 + 0.04
				elseif p1 > 0.65 then
					p1 = p1 + (p1 - 0.65)*0.4
				end
				local p1c, p2c = 24 - 21*p1, 24 - 24*p2
				e:SetSize(inv and p2c or p1c, inv and p1c or p2c)
				if pos == 1 then
					e:SetTexCoord(0.5 + 28/64*p1, 1, 0.5*p2, 0.5)
					self.spark:SetPoint("CENTER", self, "TOP", 22.5 * p1a, -22.5*p2a-1.5)
				elseif pos == 2 then
					e:SetTexCoord(0.5, 1-0.5*p2, 0.5 + 28/64*p1, 1)
					self.spark:SetPoint("CENTER", self, "RIGHT", -22.5*p2a-1.5, -22.5*p1a)
				elseif pos == 3 then
					e:SetTexCoord(0, 0.5 - 28/64*p1, 0.5, 1 - 0.5*p2)
					self.spark:SetPoint("CENTER", self, "BOTTOM", -22.5 * p1a, 1.5+22.5*p2a)
				else
					e:SetTexCoord(0.5*p2, 0.5, 0, 0.5 - 28/64*p1)
					self.spark:SetPoint("CENTER", self, "LEFT", 1.5+22.5*p2a, 22.5*p1a)
				end
				if p2 >= 0.99 then e:Hide() end
			end
		end
		local function onHide(self)
			local toExpire = GetTime() - (self.expire or 0)
			self.expire, self.pos = nil
			for i=5,9 do self[i]:Hide() end
			if -0.1 < toExpire and toExpire < 0.25 then
				self.flashAG:Play()
			end
			self:Hide()
		end
		local function onShow(self)
			self[9]:Show()
		end
		function createCooldown(parent, size)
			local cd = cc("SetScale", cc("SetAllPoints", CreateFrame("FRAME", nil, parent)), size/48)
			cc("SetScript", cc("SetScript", cc("SetScript", cd, "OnShow", onShow), "OnHide", onHide), "OnUpdate", onUpdate)
			
			cd.scale, cd.spark = size/48, cc("SetSize", cc("SetTexture", cc("SetDrawLayer", cd:CreateTexture(), "OVERLAY", 2), [[Interface\AddOns\OPie\gfx\spark]]), 24, 24)
			local sparkAG = cc("SetIgnoreFramerateThrottle", cc("SetLooping", cd.spark:CreateAnimationGroup(), "REPEAT"), true)
			cc("SetDuration", cc("SetDegrees", sparkAG:CreateAnimation("Rotation"), 90), 1/3)
			sparkAG:Play()
			
			cd.flash = cc("SetPoint", cc("SetBlendMode", cc("SetTexture", parent:CreateTexture(nil, "OVERLAY"), [[Interface\cooldown\star4]]), "ADD"), "CENTER")
			cc("SetAlpha", cc("SetSize", cd.flash, 60*size/64, 60*size/64), 0)
			cd.flashAG = cc("SetIgnoreFramerateThrottle", cd.flash:CreateAnimationGroup(), true)
			cc("SetDuration", cc("SetDegrees", cd.flashAG:CreateAnimation("ROTATION"), -90), 1/2)
			cc("SetDuration", cc("SetChange", cd.flashAG:CreateAnimation("ALPHA"), 0.7), 1/8)
			cc("SetDuration", cc("SetStartDelay", cc("SetChange", cd.flashAG:CreateAnimation("ALPHA"), -0.7), 1/8), 3/8)
			
			cd[1] = cc("SetPoint", cc("SetTexture", cd:CreateTexture(nil, "ARTWORK"), [[Interface\AddOns\OPie\gfx\borderlo]]), "BOTTOMRIGHT", cd, "RIGHT")
			cd[2] = cc("SetPoint", cc("SetTexture", cd:CreateTexture(nil, "ARTWORK"), [[Interface\AddOns\OPie\gfx\borderlo]]), "BOTTOMLEFT", cd, "BOTTOM")
			cd[3] = cc("SetPoint", cc("SetTexture", cd:CreateTexture(nil, "ARTWORK"), [[Interface\AddOns\OPie\gfx\borderlo]]), "TOPLEFT", cd, "LEFT")
			cd[4] = cc("SetPoint", cc("SetTexture", cd:CreateTexture(nil, "ARTWORK"), [[Interface\AddOns\OPie\gfx\borderlo]]), "TOPRIGHT", cd, "TOP")
			for i=1,4 do
				cd[4+i] = cc("SetPoint", cc("SetTexture", cc("SetDrawLayer", parent:CreateTexture(), "ARTWORK", 3), 1,1,1),
					(i % 4 < 2 and "TOP" or "BOTTOM") .. (i < 3 and "RIGHT" or "LEFT"), cd, "CENTER", (i < 3 and 21 or -21)*size/48, (i % 4 < 2 and 21 or -21)*size/48)
			end
			cd[9] = cc("SetTexture", cc("SetDrawLayer", parent:CreateTexture(), "ARTWORK", 3), [[Interface\AddOns\OPie\gfx\tri]])
			
			return cd
		end
	end
	
	api = {__index=api}
	function ORI_CreateDefaultIndicator(name, parent, size, ghost)
		local e = cc("SetSize", CreateFrame("Frame", name, parent), size, size)
		return setmetatable({[0]=e,
			edge = cc("SetAllPoints", cc("SetTexture", e:CreateTexture(nil, "OVERLAY"), [[Interface\AddOns\OPie\gfx\borderlo]])),
			hiEdge = cc("SetAllPoints", cc("SetTexture", cc("SetDrawLayer", e:CreateTexture(), "OVERLAY", 1), [[Interface\AddOns\OPie\gfx\borderhi]])),
			oglow = cc("SetShown", CreateQuadTexture(e, "BACKGROUND", size*2, [[Interface\AddOns\OPie\gfx\oglow]]), false),
			iglow = cc("SetAllPoints", cc("SetAlpha", cc("SetTexture", cc("SetDrawLayer", e:CreateTexture(nil), "OVERLAY", -8), [[Interface\AddOns\OPie\gfx\iglow]]), ghost and 0.60 or 1)),
			icon = cc("SetPoint", cc("SetSize", e:CreateTexture(nil, "ARTWORK"), 60*size/64, 60*size/64), "CENTER"),
			overIcon = cc("SetPoint", cc("SetDrawLayer", e:CreateTexture(), "ARTWORK", 1), "BOTTOMLEFT", e, "BOTTOMLEFT", 4, 4),
			count = cc("SetPoint", cc("SetJustifyH", e:CreateFontString(nil, "OVERLAY", "NumberFontNormalLarge"), "RIGHT"), "BOTTOMRIGHT", -4, 4),
			key = cc("SetPoint", cc("SetJustifyH", e:CreateFontString(nil, "OVERLAY", "NumberFontNormalSmallGray"), "RIGHT"), "TOPRIGHT", -1, -4),
			cdText = cc("SetPoint", e:CreateFontString(nil, "OVERLAY", "NumberFontNormalHuge"), "CENTER"),
			cd = createCooldown(e, size),
		}, api)
	end
end

local function SetAngle(self, angle, radius)
	self:SetPoint("CENTER", radius*cos(90+angle), radius*cos(angle))
end
local function SetScaleSmoothed(self, scale, speed)
	local old, limit = self:GetScale(), speed/GetFramerate();
	self:SetScale(old + min(limit, max(-limit, scale-old)));
end
local function CalculateRingRadius(n, fLength, aLength, min, baseAngle)
	if n < 2 then return min end
	local radius, mLength, astep = max(min, (fLength + aLength * (n-1))/6.2831853071796), (fLength+aLength)/2, 360 / n
	repeat
		local ox, oy, clear, angle, i = radius*cos(baseAngle), radius*sin(baseAngle), true, baseAngle + astep, 1
		while clear and i <= n do
			local nx, ny, sideLength = radius*cos(angle), radius*sin(angle), (i == 1 or i == n) and mLength or aLength
			if abs(ox - nx) < sideLength and abs(oy - ny) < sideLength then
				radius, clear = radius + 5
			end
			ox, oy, angle, i = nx, ny, angle + astep, i + 1
		end
	until clear
	return radius
end

local ORI_Slices = setmetatable({}, {__index=function(t, k) return rawset(t,k, ORI_CreateIndicator(nil, ORI_Frame, 48))[k] end})
local GhostIndication = {} do
	local spareGroups, spareSlices, currentGroups, activeGroup = {}, {}, {};
	local function freeGroup(g)
		g:Hide()
		for i=2,g.count or 0 do
			g[i]:SetShown(false)
			spareSlices[g[i]], g[i] = g[i]
		end
		spareGroups[g], g.incident, g.count = g
	end
	local function createGroup()
		return cc("SetScale", cc("SetSize", CreateFrame("Frame", nil, ORI_Frame), 1, 1), 0.80)
	end
	local function AnimateHide(self, elapsed)
		local total = ORI_ConfigCache.XTZoomTime;
		self.expire = (self.expire or total) - elapsed;
		if self.expire < 0 then
			self.expire = nil; self:SetScript("OnUpdate", nil); self:Hide();
		else
			self:SetAlpha(self.expire/total);
		end
	end
	local function AnimateShow(self, elapsed)
		local total = ORI_ConfigCache.XTZoomTime/2;
		self.expire = (self.expire or total) - elapsed;
		if self.expire < 0 then
			self.expire = nil; self:SetScript("OnUpdate", nil); self:SetAlpha(1);
		else
			self:SetAlpha(1-self.expire/total);
		end
	end
	function GhostIndication:ActivateGroup(index, count, incidentAngle, mainRadius, mainScale)
		local ret, config = currentGroups[index] or next(spareGroups) or createGroup(), ORI_ConfigCache;
		currentGroups[index], spareGroups[ret] = ret;
		if not ret:IsShown() then ret:SetScript("OnUpdate", AnimateShow); ret:Show(); end
		if activeGroup ~= ret then GhostIndication:Deactivate(); end
		if ret.incident ~= incidentAngle or ret.count ~= count then
			local radius, angleStep = CalculateRingRadius(count, 48*mainScale, 48*0.80, 30, incidentAngle-180)/0.80, 360/count;
			local angle = 90 + incidentAngle + angleStep;
			for i=2,count do
				local cell = ret[i] or next(spareSlices) or ORI_CreateIndicator(nil, ret, 48, true)
				cell:SetParent(ret); SetAngle(cell, angle, radius); cell:SetShown(true);
				ret[i], angle, spareSlices[cell] = cell, angle + angleStep;
			end
			ret.incident, ret.count = incidentAngle, count;
			ret:SetPoint("CENTER", (mainRadius/0.80+radius)*cos(incidentAngle), (mainRadius/0.80+radius)*sin(incidentAngle));
			ret:Show();
		end
		activeGroup = ret;
		return ret;
	end
	function GhostIndication:Deactivate()
		if activeGroup then
			activeGroup:SetScript("OnUpdate", AnimateHide);
			activeGroup = nil;
		end
	end
	function GhostIndication:Reset()
		for k, v in pairs(currentGroups) do
			freeGroup(v)
			currentGroups[k] = nil
		end
		activeGroup = nil
	end
	function GhostIndication:Wipe()
		GhostIndication:Reset()
		wipe(spareSlices)
	end
end

local lhc, lhcPal do
	local ht = [[ddd&ddMBLWMMML­LyL88LMLL&LhL!psF 5n&t8DqA	GD !s_aung 9G9II7$ {Oބ1C$Ift !0J)L", GD/a$}C5&!	C' 	9'yK$!Aؒ. 9/(y?/$cx0/4r!?dd=V=lK9? Ȇ,<H:#;h0:=O"?UDOL s9{I^:W} {g:n(>D-3\I "a}.z]0!щ5E [~[rh 	"̏#>G=('LɅV }$BhV!1>S/Y} +_G":A؛4:by UJ/CEX Ch#û~Ź+c= _tqo',m~9	} 	O܆I5?do$ yGnI!~? stWK(}Or1wVd GuouHPDqЏC=}%q&I$kF;AH Z"wg;+_ι EJPдws >R"ZM,S 5(yΜi? /q)G=,y !r enE?XF Gom 	$Q$ ?H.Cߴv>r/@ 7g΃;㤳#'ҳ #gh"QSğ" |7}sGq usCM>_?l# Z"H f_C'лA}l 'ܴ2לK~"'܄?_y ӔMg?Aw 	cr1'GEC\ 9c܄${1, ЈUqЋ`wFSj? euBQznI)3H K}P#_H$q/ =:^QlOOG? G%F+哯X GKw$1(ї %7#M v1ǢN~$}Ͻk-O B9_VOdqUhMH; )BI9{,{'Vg	%h@ !υ1,At >Yc[H{} cҹ?bx"	%%~ 1GQtl.?yF$ YBbLɓyؙ <9s4IU 0؟g<}RH Dܴ2YO|>JR: |EzRo( a`AjV?D_  gAE'aF ޗ;:/tKb	 Y\pϞo]1$贒r :=qa9ȿ {|'?σq W"}uwm  dODGOzr} '?63!+<zE!} !E3bA͈Yسu 1A? Yt]A}WD 	{!/7ZV{4 :HkGI/gҿ. eCG"_ozcߓ腤H~-?L]܋\I- CtЄr.e1GLa %N|GO=1' ( _JeS$;	ЄPɣ\ք}1M ֿ3?d7xM( ;%[>LFf 	"?-#G~~G"k GҺEj(}<侥 !^gI:i/Ng? "Gw}aȌ}l sMdrO]\%go /#2E.[O| [OaÑ)%Vϝq.dĒ,:$D}agJS'>U OD=I|F#8w.qЏ 	GLǉ'K Yք%("&g"ēO[}GزQF~?" Y>V92u䶑Ⱥ4#B $u?C?K?# bOs|{.g:H #&tZ\˟"ޟq?盖C O?B#9KKwp BϞ|x܈w !9	4ErйOh9П~?rN==q㻧<o9?o?YV} 9;rWyg Y֟cU#$Ig@#$|D e?1?>QʄrR#$ gȇIh:	Y;0/ȿT097!G>$HH '/XϠDH| dROЙ<!> _믰[61VZҒNW?OD ׯK9'"hPJSOuJzG/y8Z M/;2O"|]'wϟMH&߼O qs#[C&	;s3]R J	%'ٟyKO^?-	{o| 2?9z?N, ##7a&'0;kY)ik_b/}wsILVg؛o 	(._}]d=($5;⟑ҟ cCw}\F|2Su_#Q9޳a8$ :~.^idBw? -:~?Y;?s;GwΟ cЄ9l(EQt5?Lwg6zS"O;,;&ûD)9>塎 ;s'2OG/	O'ζ|Fr?"ȏu9Bi2qVhG|vE=}gzC?	0=<{՝fZ/"\%1y AJ>#b-[OoG'VMB?'%)3VDZn8J|DM5Gؿ;M}=ћ S[M܄Z#3 X<\Ϳ;ɮ$ #1;_)̏IC""E~0@gGp;c,JIbm$	'H; AUDII?E?(5Bw_~_O+rz!GK||Џ\ %I'k#[O؇h;rk (g؛uM%#Wqοǩ-r {v99,O1"?e# Zfy,<S9u #Cŋq?JWLHDq>? %;j1cDvwO!"ϧ,=Rϑro蛑 ̸#QI?s!'3&t> ;OrҾ8JZ?G ]ID{J,O9	/:u!Im%;S<s,<mq"D;$P. 	E+"YO<L":?Y,@M߿;DC!BIcЉ~Ib"GyrF<]mK~i;G?B&rB:o"J$\'Dwȿ"kUG!! G(.c&3gG~B}OPO/ 5}D#kO/δ/_9?#)"!~$OnB RGD^v>%GfwOHbFĺ·^ӟ}?J73uGBE5oٳOBX"eo&]}YKCd>?z7 sj	w1?=$4!I9r9kzĥ	]Ob]9,G1Җ|Dд:"}"ˮX5?r. IO<?Oօ]cW@[2 dG]	t1xJç$td ;һ}R}!MlD_(9ʰG$RA#!?!a9sB#A9sBP#19 B].i #Xz!)21+1?+!";B}A)BbHDDbw qu7#~y! L|IeGr\L Hח\H|D\''Z	gY	Ð#:dt&ER䒱B??zE'9MU BHğ'E<w1Wҿ?wҗ"b~t(!?3	L9ZIA	 wܒ#B		E|~NZrBs^>rt~@d}=fb>Iq?r&"G.$!/ooi{ |C5=|΍؟/t?|I$C"[ք%K+&?d <">GCH'{v?3>N.:P.	_ėb:	]JUhs:XWDSWߏ'rF?؋:j]%mYF<CiEHv E :ȗCKrot|Z>C h/[y5$#cddCO#:,C[tdYܟ%r'ct]H#EuزJ~V?~Oi=$ gF:dm윘?B	E&3΁?W!$@з8Ő@q"{zD !E'uq!NB'Kއ$?zO aMG&c/LC"D=?>Ћ^:>#:֟MRsW-8'帟w|Q|s:4[A?2 =?Ŏ&9*F}!+bGd!!os*I01IyE%fDG >J>ϼ0wr *#FEEA&Q wNI}$;eoGtnń_O'!4h? si~&;#id>ϱl}dɄmԲ>ǐwZ.CI>1H,N&ŒC$~? O?_ֲEbb㥖C?%z9=|ĺ,&ï_:?(!}<7.ƒE Zg?(x!Ϲdtd ͓Osn' w8?/z%q5GGG҃;L/gaҕxŸ9})	"ѩ1 ː'ܓ<╳;#\B~$j >ȇ[-B_w#+! %>Og_r^C?#؃&G|H&Z!,q:Qs?i< OȋE{~[sE8u1A2:I2	4}7B`As~gܒ25D~?[o؃+C&YԌ?d}uS3~D.By@%G=.*Q>' fu~E9L&y+NH !tD䅑(i'CgB%BbOR<	^(_';"	RECO?Dpv%/#B8OGؗL'B}o'?򊑇 F|e]՟93OwhH-h?c}u8	y_B})L>:?w:/M!X YgOvUh+=T}!'KG$ ]oPI>A?KuO tO&3}(3ȿyۗ/KF|)i_CB #ЇI~_>ؿ##O/Ce>k(Ĉ w8E^sAuDCICN;"~zF}dS :>Sg ?CwE:ȇ2%;Cc>O>	]lZV-$Or}}e,O@"xObi@>D.ЇU?yD~pBHhId!O,c%B!r %I'G D|q,Gv=L"TsWk3/"%;>tg>D$rK__;?T?|a>t9 !>˲It-B>.@%O$'@~3"!:*;J:WΚ?c #>'e\<Y)=^ܻo33τ _ȏ{}z__"Q߭O8'/<>Rs"$Mj} ?~2G	K?CBo; V:w:Z GO-E;O>Hs# sz#wREG:7>ybRD "?l"d2[O=b}BG/%?1ޗA	O=%ę؄otKU_{Nvԗ :>~ ?!?K:Ώ?r?ُ~D|G0ēR$C-NQʛT,˃fd󅽒?^_5׆f-S;?$L*M$Z6AŁ̓I6@rdL3Gx?Pܙ՛`6o)=)YړUnoƝb؋1CÕ])+ܟ<@ܞ;,@Iڈ&Ӕ̌~|C`瀅LW*):<#k򐁙Y=kRAC:Y^'v]T'&b?XŁ̀ђomU	w4<&=Ο:-s=T)k~u'PY߁`8PIf=ΔLHҁ͙Oȝ&ğ:=HHTlp?ao]ЕBd/ㄋ%	)$O9<Yѕ=.$ϐ<d򎁌=̒-&5܇.-'-鄡%Zd=ΔB<=֙n-i<M6Q򜠊-dņVY-o>1d5=Q	N$Ln܎PP?~ܑWA=Q&H<rB5~)Ӟ.̖֚Cˈ&}Gs<<[Z1	Иu'*Ŏ#=ΟM6%tNE"<%7'M1	InLNi?sԈtNI<vh 9rs_)p-ۑڊPG$蟓.QEu6-jjou䊎dr<瑠=o9=>s,TUdxEu=Ɗ-䉤ތ%43؞&r>?l^>Rސ?ތ蒜+	Tj)(W<;=)T:.ȃʘT/tYZ>1ܖ)(>5PWsĔ,ύ5ēq513Bl2ň>1Y>1$CRlg$<HO5EYfۘyc҂moo8<獫׍)0ޘo֜m%<%hG@j{A5n,d-[<U+E'T6\VcLT-n)؆/f1yč=]̈&X:}-uYn3=J?)ɛ6ÞN㕏;oC5ʏЍ)mtŀЖ9(&VT$j[YrQ;́_Uv*usx*=L@eT%o<Hzz*H6{?Ւ-?{-|ߋɝ3τT?􏑚>R)tpo9mL<ˍ&9E#ؖ<v--K%,'z:}=3͝]ύW	&1H?םٓMWڋ/u{k,䒲׃d|_06=1wlB~HNėYdҞđBˎ=/50ܓ<;N%=ĄÊގ]??S3A&&EZs)돸zzOčoֈړY%_U5hY߅o[)4>sŉƍ_<2ѐ=dp%.c=kOQꚧ5*ڒ[?ވ^g/I-oEIh=ad>1É@0'6d<'mz6*(C	;%TI@3[<᝔z!-(}9=)"ށuӐo._1#˂5Z=CJÛGpL>5횿;=ܞXߘTg'*E</xw՗1-Ɖޟ:=ΗߙY]1-&?>sA-_=)=-ڞ됗랬&Nta>s(؇Pr1Ԉۃ:O!{*)񛾂~(uCM=kIP>s=rᙪNa˛N->RƊl)(wf#-$QZt>1`芺0 u⚛oӉ._ႉf%č㝺%ox|a$⋱k,?66#_Axg,oXC4<]3T⒜Ќ~X+N˔؃>s}=kW͖ğ=;Y&zA_]wi$<+6րALʀC2ЎQ"ΝQ $=Čj={oz?᚛?Y]/]j[xIۛ<6Ɯ)3eˆoZ?τq	c,5邵d&+ލAd]turu<E+Š*܌<B&.Aឬl|8]Ō[ـ?1ޘ9ɁI9P$Z:Ar/m,=kq&g=o*=ʛ<~ĐX&mu捷Uˁ11EːfޕB&|۝A)ޙz-Bה,٠6цW{ֆW@=)@+ןoО׆-/=΍A?]Ō=6vxcʏۉ*;<ۆ$31ȄّBL*&Ag{ߠB=ƉVܛ$N6ȕzWƔ@̎=)}YH5Ō=l͚x$L&WLے $ӌ=,=*ӏ51Oڎ/X׋,*&ϝʕ(/~u+˛ԋ?|&Id]]
	local byte, cr, k2, a3, b3, m3, sp = string.byte, {}
	for i=1,byte(ht,1) do cr[3*i-2], cr[3*i-1], cr[3*i] = byte(ht, 3*i-1, 3*i+1) end
	for i=1,#cr do cr[i] = cr[i] > 1 and (cr[i]/255) or 0 end
	local function r16(p) local a,b = byte(ht, p, p+1) return a*256 + b end	
	k2, m3, a3, b3, sp, lhcPal = r16(#cr+2), r16(#cr+4), r16(#cr+6), r16(#cr+8), #cr + 10, cr

	local function get(s)
		local h, l, a, a2, b, c, d, e, f, o, m = 8388593, #s
		for i=1, l+1, 2 do
			a, b = byte(s, i, i+1)
			h = (h * 4093 + (a or l) * k2 + (b or l) * 65521) % 268435399
		end
		m = -2 * (((h * a3 + b3) % 2147483629) % m3 + 1)
		a, b = byte(ht, m, m + 1)
		if a < 128 then
			return ((h * (a - a % 4) / 4) % 2147483629 % 2 > 0 and (b % 32) or ((a % 4) * 8 + (b - b % 32) / 32)) * 3 - 2
		else
			o = sp + a*256 + b - 32769
			m, a = byte(ht, o, o + 1)
			a = (h * (m * 256 + a) + m) % 2147483629 % m
			a, a2 = (a - a % 8)*5/8, a % 8
			b, c, d, e, f = byte(ht, o+a+2, o+a+6)
			a = b * 4294967296 + c * 16777216 + d * 65536 + e * 256 + f
			return ((a - a % 32^a2) / 32^a2 % 32) * 3 - 2
		end
	end
	lhc = setmetatable({}, {__index=function(t, s)
		local i = type(s) == "string" and get(s:match("[^/\\]*$"):lower():gsub("%.blp$", ""))
		if s then t[s] = cr[i] and i or -3 end
		return i
	end})
end

local ORI, ORI_cR, ORI_cG, ORI_cB, ORI_caption, ORI_icon, ORI_qHint = {}, {}, {}, {}, {}, {}, {}
local function ORI_SliceColor(token, icon)
	if ORI_cR[token] then return ORI_cR[token], ORI_cG[token], ORI_cB[token] end
	local li = lhc[icon] or -3
	return lhcPal[li] or 0.70, lhcPal[li+1] or 1, lhcPal[li+2] or 0.6
end
local function extractAux(ext, v)
	if not ext then
	elseif v == "coord" and type(ext.iconCoords) == "table" then
		return unpack(ext.iconCoords)
	elseif v == "coord" then
		return ext.iconCoords()
	elseif v == "color" and type(ext.iconR) == "number" and type(ext.iconG) == "number" and type(ext.iconB) == "number" then
		return ext.iconR, ext.iconG, ext.iconB
	end
end
local function check(ok, ...)
	if ok then return ... end
end
local function ORI_UpdateCenterIndication(self, si, osi)
	local tok, usable, state, icon, caption, count, cd, cd2, tipFunc, tipArg, ext = OneRingLib:GetOpenRingSliceAction(si);
	caption = ORI_caption[tok] or caption
	ORI_CenterCooldownText:SetFormattedText(cooldownFormat(tok and ORI_ConfigCache[usable and "ShowRecharge" or "ShowCooldowns"] and cd or 0));
	ORI_CenterCaption:SetText(tok and ORI_ConfigCache.ShowCenterCaption and caption or "");
		
	if tok then
		local r,g,b = ORI_SliceColor(tok, ORI_icon[tok] or icon or "INV_Misc_QuestionMark")
		ORI_Pointer:SetVertexColor(r,g,b, 0.9);
		ORI_Circle:SetVertexColor(r,g,b, 0.9);
		ORI_Glow:SetVertexColor(r,g,b);
		ORI_CenterCaption:SetTextColor(r,g,b);
		ORI_CenterCooldownText:SetTextColor(r,g,b);
	elseif si ~= osi then
		ORI_Pointer:SetVertexColor(1,1,1, 0.1);
		ORI_Circle:SetVertexColor(1,1,1, 0.3);
		ORI_Glow:SetVertexColor(0.75,0.75,0.75);
	end

	if ORI_ConfigCache.UseGameTooltip then
		if tipFunc and tipArg then
			GameTooltip_SetDefaultAnchor(GameTooltip, ORI_Frame)
			tipFunc(GameTooltip, tipArg)
			GameTooltip:Show()
		elseif caption and caption ~= "" then
			GameTooltip_SetDefaultAnchor(GameTooltip, ORI_Frame)
			GameTooltip:AddLine(caption)
			GameTooltip:Show()
		elseif GameTooltip:IsOwned(ORI_Frame) then
			GameTooltip:Hide()
		end
	end

	local gAnim, gEnd, oIG, time, usable = self.gAnim, self.gEnd, self.oldIsGlowing, GetTime(), not not usable
	if usable ~= oIG then
		gAnim, gEnd = usable and "in" or "out",  time + 0.3 - (gEnd and gEnd > time and (gEnd-time) or 0);
		self.oldIsGlowing, self.gAnim, self.gEnd = usable, gAnim, gEnd;
		ORI_Glow:SetShown(true);
	end
	if gAnim and gEnd <= time or oIG == nil then
		self.gAnim, self.gEnd = nil, nil;
		ORI_Glow:SetShown(usable);
		ORI_Glow:SetAlpha(0.75);
	elseif gAnim then
		local pg = (gEnd-time)/0.3*0.75;
		ORI_Glow:SetAlpha(usable and (0.75 - pg) or pg);
	end
	self.oldSlice = si
end
local function ORMI_UpdateSlice(indic, selected, tok, usable, state, icon, _, count, cd, cd2, tf, ta, ext)
	state, icon, ext = state or 0, ORI_icon[tok] or icon  or "Interface/Icons/INV_Misc_QuestionMark", not ORI_icon[tok] and ext or nil
	local active, overlay, faded, usableCharge = state % 2 > 0, state % 4 > 1, not usable, usable or (state % 128 >= 64)
	local c, ok, r,g,b = faded and 0.5 or 1
	indic:SetIcon(icon)
	if ext then
		indic:SetIconTexCoord(check(pcall(extractAux, ext, "coord")))
		ok, r,g,b = pcall(extractAux, ext, "color", c)
	end
	indic:SetIconVertexColor(r or c, g or c, b or c)
	indic:SetDominantColor(ORI_SliceColor(tok, icon))
	indic:SetOuterGlow(overlay)
	indic:SetOverlayIcon((ORI_qHint[tok] or ((state or 0) % 64 >= 32)) and "Interface\\MINIMAP\\TRACKING\\OBJECTICONS", 21, 28, 40/256, 64/256, 32/64, 1)
	indic:SetCooldown(cd, cd2, usableCharge)
	indic:SetCooldownFormattedText(cooldownFormat(ORI_ConfigCache[usableCharge and "ShowRecharge" or "ShowCooldowns"] and cd or 0))
	indic:SetCount((count or 0) > 0 and count)
	indic:SetActive(active)
	indic:SetHighlighted(selected and not faded)
end
local function ORI_GhostUpdate(self, slice)
	local count, offset = self.count, self.offset;
	local _, _, _, ghostCount = OneRingLib:GetOpenRingSlice(slice or 0);
	if (ghostCount or 0) == 0 then return GhostIndication:Deactivate(); end
	local scaleM = ORI_ConfigCache.MIScale and 1.10 or 1;
	local group = GhostIndication:ActivateGroup(slice, ghostCount, 90 - 360/count*(slice-1) - offset, self.radius*scaleM, 1.10);
	for i=2,ghostCount do
		ORMI_UpdateSlice(group[i], false, OneRingLib:GetOpenRingSliceAction(slice, i));
	end
end
local function ORI_Update(self, elapsed)
	local time, config, count, offset = GetTime(), ORI_ConfigCache, self.count, self.offset

	local scale, l, b, w, h = self:GetEffectiveScale(), self:GetRect();
	local x, y = GetCursorPosition();
	local dx, dy = (x / scale) - (l + w / 2), (y / scale) - (b + h / 2);
	local radius2 = dx*dx+dy*dy;

	local angle, isInFastClick = atan2(dy, dx) % 360, config.CenterAction and radius2 <= 400 and self.fastClickSlice > 0 and self.fastClickSlice <= self.count;
	if isInFastClick then
		angle = (offset + 90 - (self.fastClickSlice-1)*360/count) % 360;
	end

	local oangle = (not isInFastClick) and self.angle or angle;
	local adiff, arate = min((angle-oangle) % 360, (oangle-angle) % 360);
	if adiff > 60 then
		arate = 420 + 120*sin(min(90, adiff-60));
	elseif adiff > 15 then
		arate = 180 + 240*sin(min(90, max((adiff-15)*2, 0)));
	else
		arate = 20 + 160*sin(min(90, adiff*6));
	end
	local abound, arotDirection = arate/GetFramerate(), ((oangle - angle) % 360 < (angle - oangle) % 360) and -1 or 1;
	abound = abound * 2^config.XTPointerSpeed;
	self.angle = (adiff < abound) and angle or (oangle + arotDirection * abound) % 360;
	ORI_Pointer:SetRotation(self.angle/180*3.1415926535898 - 90/180*3.1415926535898);

	local si = isInFastClick and self.fastClickSlice or (count <= 0 and 0) or (radius2 < 1600 and 0) or
		(floor(((atan2(dx, dy) - offset) * count/360 + 0.5) % count) + 1)
	ORI_UpdateCenterIndication(self, si, self.oldSlice);

	if config.MultiIndication and count > 0 then
		local cmState, mut = (IsShiftKeyDown() and 1 or 0) + (IsControlKeyDown() and 2 or 0) + (IsAltKeyDown() and 4 or 0), self.schedMultiUpdate or 0;
		if self.omState ~= cmState or mut >= 0  then
			self.omState, self.schedMultiUpdate = cmState, -0.05;
			for i=1,count do
				ORMI_UpdateSlice(ORI_Slices[i], si == i, OneRingLib:GetOpenRingSliceAction(i));
			end
			if config.GhostMIRings then
				ORI_GhostUpdate(self, si);
			end
		else
			self.schedMultiUpdate = mut + elapsed;
		end

		for i=1,config.MIScale and count or 0 do
			SetScaleSmoothed(ORI_Slices[i], i == si and 1.10 or 1, 2^ORI_ConfigCache.XTScaleSpeed)
		end
	end
end
local function ORI_ZoomIn(self, elapsed)
	self.eleft = self.eleft - elapsed;
	local delta, config = max(0, self.eleft/ORI_ConfigCache.XTZoomTime), ORI_ConfigCache;
	if delta == 0 then self:SetScript("OnUpdate", ORI_Update); end
	self:SetScale(config.RingScale/max(0.2,cos(65*delta))); self:SetAlpha(1-delta);
	return ORI_Update(self, elapsed);
end
local function ORI_ZoomOut(self, elapsed)
	self.eleft = self.eleft - elapsed;
	local delta = max(0, self.eleft/ORI_ConfigCache.XTZoomTime);
	if delta == 0 then return self:Hide(), self:SetScript("OnUpdate", nil); end
	if ORI_ConfigCache.MultiIndication and ORI_ConfigCache.MISpinOnHide then
		local count = self.count;
		if count > 0 then
			local baseAngle, angleStep, radius, prog = 45 - self.offset + 45*delta, 360/count, self.radius, (1-delta)*150*max(0.5, min(1, GetFramerate()/60));
			for i=1,count do
				ORI_Slices[i]:SetPoint("CENTER", cos(baseAngle)*radius + cos(baseAngle-90)*prog, sin(baseAngle)*radius + sin(baseAngle-90)*prog);
				baseAngle = baseAngle - angleStep;
			end
		end
		self:SetScale(ORI_ConfigCache.RingScale*(1.75 - .75*delta));
	else
		self:SetScale(ORI_ConfigCache.RingScale*delta);
	end
	self:SetAlpha(delta);
end
ORI_Frame:SetScript("OnHide", function(self)
	if self:IsShown() and self:GetScript("OnUpdate") == ORI_ZoomOut then
		self:SetScript("OnUpdate", nil)
		self:Hide()
	end
end)

function ORI:Show(ringName, fcSlice, fastOpen)
	local frame, config, _ = ORI_Frame, ORI_ConfigCache;
	_, frame.count, frame.offset = OneRingLib:GetOpenRing(config);
	frame.radius = CalculateRingRadius(frame.count or 3, 48, 48, 95, 90-(frame.offset or 0))
	frame:SetScript("OnUpdate", ORI_ZoomIn);
	frame.eleft, frame.fastClickSlice, frame.oldSlice, frame.angle, frame.omState, frame.oldIsGlowing = config.XTZoomTime * (fastOpen and 0.5 or 1), fcSlice or 0, -1;
	ORI_SetRotationPeriod(config.XTRotationPeriod)
	MouselookStop();
	GhostIndication:Reset();

	local astep, radius = frame.count == 0 and 0 or -360/frame.count, frame.radius;
	for i=1,config.MultiIndication and frame.count or 0 do
		local indic, _, _, sliceBind  = ORI_Slices[i], OneRingLib:GetOpenRingSlice(i);
		indic:SetBindingText(config.ShowKeys and sliceBind and shortBindName(sliceBind));
		SetAngle(indic, (i - 1) * astep - frame.offset, radius);
		indic:SetShown(true)
		indic:SetScale(1)
	end
	for i=config.MultiIndication and (frame.count+1) or 1, #ORI_Slices do
		ORI_Slices[i]:SetShown(false)
	end

	config.RingScale = max(0.1, config.RingScale);
	frame:SetScale(config.RingScale); frame.anchor:SetScale(config.RingScale);
	if config.RingAtMouse then
		local es, cx, cy = frame:GetEffectiveScale(), GetCursorPosition()
		frame.anchor:SetPoint("CENTER", nil, "BOTTOMLEFT", (cx + config.IndicationOffsetX)/es, (cy - config.IndicationOffsetY)/es);
	else
		local es = frame:GetEffectiveScale()
		frame.anchor:SetPoint("CENTER", nil, "CENTER", config.IndicationOffsetX/es, -config.IndicationOffsetY/es);
	end
	frame:Show();

	ORI_Update(frame, 0);
end
function ORI:Hide()
	ORI_Frame:SetScript("OnUpdate", ORI_ZoomOut);
	ORI_Frame.eleft = ORI_ConfigCache.XTZoomTime;
	GhostIndication:Deactivate();
	if GameTooltip:IsOwned(ORI_Frame) then
		GameTooltip:Hide()
	end
end
function ORI:SetDisplayOptions(token, icon, caption, r,g,b)
	if type(r) ~= "number" or type(g) ~= "number" or type(b) ~= "number" then r,g,b = nil end
	ORI_cR[token], ORI_cG[token], ORI_cB[token], ORI_caption[token], ORI_icon[token] = r,g,b, caption, icon
end
function ORI:SetQuestHint(sliceToken, hint)
	ORI_qHint[sliceToken] = hint or nil
end
function ORI:GetTexColor(icon)
	return ORI_SliceColor(nil, icon)
end
function ORI:SetIndicatorConstructor(func)
	ORI_CreateIndicator = func or ORI_CreateDefaultIndicator
	GhostIndication:Wipe()
	for k,v in pairs(ORI_Slices) do
		ORI_Slices[k] = nil
		v:SetShown(false)
	end
	ORI_Frame:Hide()
end

OneRingLib.ext.OPieUI = ORI
ORI:SetIndicatorConstructor(ORI_CreateDefaultIndicator)
OneRingLib:SetAnimator(ORI)
for k,v in pairs({ShowCenterCaption=false, ShowCooldowns=false, ShowRecharge=false, MultiIndication=true, UseGameTooltip=true, ShowKeys=true,
	MIScale=true, MISpinOnHide=true, GhostMIRings=true, XTPointerSpeed=0, XTScaleSpeed=0, XTZoomTime=0.3, XTRotationPeriod=4}) do
	OneRingLib:RegisterOption(k,v)
end